/*
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Copyright 2005 TU-Berlin, Tilman Knebel tk@ni.cs.tu-berlin.de
*/
#include <stdio.h>
#include <string.h>
#include "mex.h"
#include "psvm.h"


/* mexFunction is the gateway routine for the MEX-file. */ 
void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	//nlhs - nr of left hand side args (output)
	//plhs - left hand side args (output)
	//nrhs - nr of right hand side args (input)
	//prhs - right hand side args (input)
	//
	//[K]=psvmkernel(X,Z,[kerneltype p1 p2 p3]
	//check type of input args
	const mxArray *mxX,*mxZ,*mxKpar;
	mxX=prhs[0];
	mxZ=prhs[1];
	mxKpar=prhs[2];

	if(nrhs!=3 || nlhs>1)
		mexErrMsgTxt("usage: K=psvmkernel(X,Z,[kerneltype p1 p2 p3])\n"
		"or     K=psvmkernel(X,X,[kerneltype p1 p2 p3])");
	if(!mxIsNumeric(mxX) || !mxIsNumeric(mxZ) || !mxIsNumeric(mxKpar))
		mexErrMsgTxt("all input args must be numeric");
	if(!mxIsDouble(mxX) || !mxIsDouble(mxZ) || !mxIsDouble(mxKpar))
		mexErrMsgTxt("all input args must be double type");
	if(mxIsComplex(mxX) || mxIsComplex(mxZ) || mxIsComplex(mxKpar))
		mexErrMsgTxt("complex data not supported");

	//check dimension of input args
	int dimX,dimZ,dimKpar;
	const int *pdimX,*pdimZ,*pdimKpar;//pointer to number of indices
	int L,P,N,pars;//number of indices per dimension
	dimX=mxGetNumberOfDimensions(mxX);
	dimZ=mxGetNumberOfDimensions(mxZ);
	dimKpar=mxGetNumberOfDimensions(mxKpar);
	pdimX=mxGetDimensions(mxX);
	pdimZ=mxGetDimensions(mxZ);
	pdimKpar=mxGetDimensions(mxKpar);
	if(dimX!=2)
		mexErrMsgTxt("X must be a two dimensional matrix\n");
	if(dimZ!=2)
		mexErrMsgTxt("Z must be a two dimensional matrix\n");
	if(dimKpar!=2 || (pdimKpar[1]!=1 && pdimKpar[0]!=1))
		mexErrMsgTxt("Parameters must be a vector\n");

	//check size of input args
	L=pdimX[0];
	N=pdimX[1];
	P=pdimZ[0];
	if(pdimZ[1]!=N)
		mexErrMsgTxt("X and Z must have same number of cols\n");
	pars=max(pdimKpar[0],pdimKpar[1]);

	PSVMmodel psvm;
	//put third arg into psvm structure, kernel parameter
	double* aX,*aZ,*aKpar;
	aX=mxGetPr(mxX);
	aZ=mxGetPr(mxZ);
	aKpar=mxGetPr(mxKpar);

	if(pars>=1)psvm.par.kernel=(int)aKpar[0];
	if(pars>=2)psvm.par.k1=(double)aKpar[1];
	if(pars>=3)psvm.par.k2=(double)aKpar[2];
	if(pars>=4)psvm.par.k3=(double)aKpar[3];

	int p,l,n;
	double* aK;
	plhs[0]=mxCreateDoubleMatrix(L,P,mxREAL);
	aK=mxGetPr(plhs[0]);

	int pL,pL_l,nL_l,nP_p;
	double d;
	if(psvm.par.kernel==LINEAR)
	{
		pL=0;
		for(p=0;p<P;p++){
			pL_l=pL;
			for(l=0;l<L;l++)
            {
				d=0;
				nL_l=l;nP_p=p;
				for(n=0;n<N;n++)
				{
					d+=aX[nL_l]*aZ[nP_p];
					nL_l+=L;nP_p+=P;
				}
				aK[pL_l]=psvm.kf_LINEAR(d);
				pL_l++;
			}
			pL+=L;
		}
	}else if(psvm.par.kernel==POLY)
	{
		pL=0;
		for(p=0;p<P;p++){
			pL_l=pL;
			for(l=0;l<L;l++)
            {
				d=0;
				nL_l=l;nP_p=p;
				for(n=0;n<N;n++)
				{
					d+=aX[nL_l]*aZ[nP_p];
					nL_l+=L;nP_p+=P;
				}
				aK[pL_l]=psvm.kf_POLY(d);
				pL_l++;
			}
			pL+=L;
		}
	}else if(psvm.par.kernel==RBF)
	{
		pL=0;
		for(p=0;p<P;p++){
			pL_l=pL;
			for(l=0;l<L;l++)
            {
				d=0;
				nL_l=l;nP_p=p;
				for(n=0;n<N;n++)
				{
					d+=sqr(aX[nL_l]-aZ[nP_p]);
					nL_l+=L;nP_p+=P;
				}
				aK[pL_l]=psvm.kf_RBF(d);
				pL_l++;
			}
			pL+=L;
		}
	}else if(psvm.par.kernel==SIGMOID)
	{
		pL=0;
		for(p=0;p<P;p++){
			pL_l=pL;
			for(l=0;l<L;l++)
            {
				d=0;
				nL_l=l;nP_p=p;
				for(n=0;n<N;n++)
				{
					d+=aX[nL_l]*aZ[nP_p];
					nL_l+=L;nP_p+=P;
				}
				aK[pL_l]=psvm.kf_SIGMOID(d);
				pL_l++;
			}
			pL+=L;
		}
	}else if(psvm.par.kernel==SIN)
	{
		pL=0;
		for(p=0;p<P;p++){
			pL_l=pL;
			for(l=0;l<L;l++)
            {
				d=0;
				nL_l=l;nP_p=p;
				for(n=0;n<N;n++)
				{
					d+=sqr(aX[nL_l]-aZ[nP_p]);
					nL_l+=L;nP_p+=P;
				}
				aK[pL_l]=psvm.kf_SIN(sqrt(d));
				pL_l++;
			}
			pL+=L;
		}
	}else if(psvm.par.kernel==PLUMMER)
	{
		pL=0;
		for(p=0;p<P;p++){
			pL_l=pL;
			for(l=0;l<L;l++)
            {
				d=0;
				nL_l=l;nP_p=p;
				for(n=0;n<N;n++)
				{
					d+=sqr(aX[nL_l]-aZ[nP_p]);
					nL_l+=L;nP_p+=P;
				}
				aK[pL_l]=psvm.kf_PLUMMER(d);
				pL_l++;
			}
			pL+=L;
		}
	}else if(psvm.par.kernel==USER)
	{
		pL=0;
		for(p=0;p<P;p++){
			pL_l=pL;
			for(l=0;l<L;l++)
            {
				d=0;
				nL_l=l;nP_p=p;
				for(n=0;n<N;n++)
				{
					d+=aX[nL_l]*aZ[nP_p];
					nL_l+=L;nP_p+=P;
				}
				aK[pL_l]=psvm.kf_USER(d);
				pL_l++;
			}
			pL+=L;
		}
	}
}

