/*
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Copyright 2005 TU-Berlin, Tilman Knebel tk@ni.cs.tu-berlin.de
*/
#include <stdio.h>
#include <string.h>
#include "mex.h"
#include "psvm.h"


/* mexFunction is the gateway routine for the MEX-file. */ 
void mexFunction(int nlhs, mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
	verbose=false;
	//nlhs - nr of left hand side args (output)
	//plhs - left hand side args (output)
	//nrhs - nr of right hand side args (input)
	//prhs - right hand side args (input)

	//[a]=psvmsmo(K,Y,[C epsilon epsi_zero uppertol timeout algo e2 e3])

	const mxArray *mxK,*mxY,*mxppar;
	mxK=prhs[0];
	mxY=prhs[1];
	mxppar=prhs[2];
	//check type of input args
	if(nrhs!=3 || nlhs>2)
		mexErrMsgTxt("usage: a=psvmsmo(K,Y,[C epsilon epsi_zero uppertol timeout algo e2 e3])\n"
		"[a,diag]=psvmsmo(K,Y,[C epsilon epsi_zero uppertol timeout algo e2 e3])\n"
		"diag=[time,qcount,paircount,blockcount]\n");

	if(!mxIsNumeric(mxK) || !mxIsNumeric(mxY) || !mxIsNumeric(mxppar))
		mexErrMsgTxt("all input args must be numeric");
	if(!mxIsDouble(mxK) || !mxIsDouble(mxY) || !mxIsDouble(mxppar))
		mexErrMsgTxt("all input args must be double type");
	if(mxIsComplex(mxK) || mxIsComplex(mxY) || mxIsComplex(mxppar))
		mexErrMsgTxt("complex data not supported");

	//check dimension of input args
	int dimK,dimY,dimppar;
	const int *pdimK,*pdimY,*pdimppar;
	double* aK,*aY,*appar;
	int L,P,ppars,Ys;//input
	int features;//output
	dimK=mxGetNumberOfDimensions(mxK);
	dimY=mxGetNumberOfDimensions(mxY);
	dimppar=mxGetNumberOfDimensions(mxppar);
	pdimK=mxGetDimensions(mxK);
	pdimY=mxGetDimensions(mxY);
	pdimppar=mxGetDimensions(mxppar);
	if(dimK!=2)
		mexErrMsgTxt("first arg: K must be a two dimensional array");
	if(dimY!=2 || (pdimY[1]!=1 && pdimY[0]!=1))
		mexErrMsgTxt("second arg: Y must be a vector\n");
	if(dimppar!=2 || (pdimppar[1]!=1 && pdimppar[0]!=1))
		mexErrMsgTxt("third arg: Parameters must be a vector\n");

	L=pdimK[0];//rows in K-matrix
	P=pdimK[1];//cols in K-matrix
	ppars=max(pdimppar[0],pdimppar[1]);//parameter vector
	Ys=max(pdimY[0],pdimY[1]);//number of Y elements

	//check size of input args
	if(Ys!=L)
		mexErrMsgTxt("Y must have same number of elements as rows in K\n");
	aK=mxGetPr(mxK);
	aY=mxGetPr(mxY);
	appar=mxGetPr(mxppar);

	PSVMmodel psvm;//create psvm

	//read psvmpars
	if(ppars>=1)psvm.par.C=(ftype)appar[0];
	if(ppars>=2)psvm.par.epsi=(ftype)appar[1];
	if(ppars>=3)psvm.par.epsiTol=(ftype)appar[2];
	if(ppars>=4)psvm.par.timeout=(int)appar[3];
	if(ppars>=5)psvm.par.doreduction=(int)appar[4];
	if(ppars>=6)psvm.par.redsize=(int)appar[5];
	if(ppars>=7)psvm.par.redfact=(double)appar[6];
	if(ppars>=8)psvm.par.domaxselection=(int)appar[7];
	if(ppars>=9)psvm.par.selectionfact=(double)appar[8];
	if(ppars>=10)psvm.par.doblock=(int)appar[9];
	if(ppars>=11)psvm.par.blockNmax=(int)appar[10];
	if(ppars>=12)psvm.par.blockCmax=(int)appar[11];
	if(ppars>=13)psvm.par.blockOver=(double)appar[12];
	if(ppars>=14)psvm.par.doannealing=(int)appar[13];
	if(ppars>=15)psvm.par.estart=(ftype)appar[14];
	if(ppars>=16)psvm.par.edec=(ftype)appar[15];
	if(ppars>=17)psvm.par.eterm=(ftype)appar[16];

	//reserve space for input K(ernel)-matrix in psvm structure
	Dataset set;
	set.reserve(L,P,true);
	set.mDim=P;
	set.mN=L;

	//get K into psvm structure 'set'
	int all=L*P;
	for(int i=0;i<all;i++)
		set.mData[i]=(ftype)aK[i];
	set.mTrans=true;

	//put input Y-data in psvm structure
	for(int i=0;i<L;i++)
		set.mY[i]=(ftype)aY[i];

	if(!psvm.getRankFromSet(&set))
		mexPrintf("SMO skipped after timeout\n");
	features=psvm.rank.mFeat.size();

	//create sparse destination matrix for alpha
	plhs[0]=mxCreateSparse(P,1,features,mxREAL);
	double* aa;
	int* rr;
	aa=mxGetPr(plhs[0]);//get pointer to vector data
	rr=mxGetIr(plhs[0]);//get pointer to references
	for(int i=0;i<features;i++)
	{
		aa[i]=psvm.rank.mFeat[i].a;
		rr[i]=psvm.rank.mFeat[i].dimRef;
	}
	mxGetJc(plhs[0])[0]=0;//fill dimension information
	mxGetJc(plhs[0])[1]=features;
	if(nlhs==2)
	{
		plhs[1]=mxCreateDoubleMatrix(4,1,mxREAL);
		mxGetPr(plhs[1])[0]=(double)psvm.smo.mTCount;
		mxGetPr(plhs[1])[1]=(double)psvm.smo.mQCount;
		mxGetPr(plhs[1])[2]=(double)psvm.smo.mACount;
		mxGetPr(plhs[1])[3]=(double)psvm.smo.mBCount;
	}
	set.free();
}

