/*
Software library for offering Dataset and P-SVM functionality according to

"Sepp Hochreiter, Klaus Obermayer;
 Support Vector Machines for Dyadic Data, Neural Computation, 2005"

Copyright 2005 TU-Berlin, Tilman Knebel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

If you use the software please cite 
"Sepp Hochreiter, Klaus Obermayer;
 Support Vector Machines for Dyadic Data, Neural Computation, 2005"
in your publications.
*/
#include <cstring>
#include <cmath>
#include <cstdio>
#include <ctime>
#include <cstdlib>
#include <list>
#include <algorithm>
#include <vector>
#include <float.h>
#include <stdarg.h>
using namespace std;
//#define smofloat //if smofloat is not defined, precision is double
//smofloat should be a precompiler definition if float precision reaches out

#if defined(smofloat)
#define ftype float //reduced internal precision 
#else 
#define ftype double//works better
#endif

//math functions:
template<class T> inline T sqr(T x) {return x*x;};
template<class T> inline signed char signc(T x) {if (x>0) return(1); else return(-1);};
double timeDiff(clock_t t0,clock_t tk);

enum eKernelType {LINEAR=0, POLY=1, RBF=2, SIGMOID=3, PLUMMER=4, SIN=5, USER=6};	// kernel types 

//class members labeled with leading "m"

//parameter Structure
struct Param
{//PSVM parameter list
	Param()
	{//initializer
		//basic parameters for SMO:
		C=1e9;//regularization
		epsi=(ftype)-1;//correlation coefficient
		epsiTol=(ftype)0.05;//termination tolerance factor
#if defined(smofloat)
		singular=0.9999;//nearly 1, watch precision
#else
		singular=0.9999999999;//nearly 1, watch precision
#endif
		timeout=-1;//-1 = no timeout,>0 - timeout after seconds

		//heuristics for speedup
		doannealing=1;//epsilon annealing
		estart=(ftype)0.05;//start with epsilon=estart*max(mXY)
		edec=(ftype)0.95;//decrease epsilon by this factor
		eterm=4;//terminate with epsiTol*etern

		doblock=1;//block optimization
		blockNmax=21;//begin operation with at least 'blockNmax' features, or
		blockCmax=3;//maximum of alphas on bound during block optimization 
		blockOver=3;//begin operation after at least 'blockOver'*'current blocksize' iterations

		domaxselection=1;//1 - use maximum of KKT violation, 0 - chose index randomly, 2 - with violation above selectionfact*maxviolation, 3 - maximum delta alpha constrained, 4 - maximum delta alpha constrained exact signum
		selectionfact=0.5;//only for mode 2

		doreduction=0;//reduce problem
		redsize=50;//lower is softer,
		redfact=5;//epsilon modification for solving subsets, lower is softer

		//parameters for psvm and loops
		kernel=LINEAR;k1=0.1;k2=0.1;k3=2;
		isrelation=false;//true - use input data as kernel matrix
		shuffle=1;//0 - no shuffle of datasets, 1 - shuffle deterministic, 2 - shuffle with timeseed
		flimit=-1;//limit features after training, not deterministic with low C
		allowCcut=0;//0 - if more features bounded than flimit, delete model
				//useful in modelselection to prevent too complex models in deterministic way
				//1 - cut all but the first flimit features, even if they are all bounded and not sortable
		infold=0;//inner crossvalidation for feature ranking, -1=leave-one-out 0,1=no fold >=2 tile in subsets
		outfold=0;//outer crossvalidation for GE-estimation, -1=leave-one-out 0,1=no fold >=2 tile in subsets
		//outfold currently not used
		ms_e1=epsi;//hyperparameter selection
		ms_e2=epsi;
		ms_ed=1;
		ms_c1=C;
		ms_c2=C;
		ms_cd=1;

		doclass=true;//important for choosing the final model after modelselection
		doregress=true;//if both are true, regression is dominant

		doretrain=false;//retraining for recalculation of features
		retrain_fe=0.1;//factor for epsilon 

		dobalanceY=false;//use bias removal function
		useBalancedErrors=false;//use balaned errors instead of simple error rate

		useNormLimit=false;
	};
	ftype C;//upper limit alpha
	ftype epsi;//noise compensation, a low amount leads to a more complex model (watch computational time)
	ftype epsiTol;//tolerance epsilon against upper bound C, too low values will internally be corrected
	double singular;//limit against division by zero, should be a bit lower than one
	int timeout;//maximum number of seconds running the smo (-1=no time limit)

	int kernel;//kernel type, see enum 
	double k1,k2,k3;//kernel parameter, k1=factor, k2=offset, k3=exponent
	bool isrelation;//SVM is in relational mode, input data is interpreted as kernel matrix
	int shuffle;//0 - no shuffle, 1 - shuffle with constant seed, 2 - shuffle with time seed

	//heuristics for speedup
	int doannealing;//0 - no epsilon annealing, 1 - do epsilon annealing
	ftype estart;//relative starting factor(1=highest 0=lowest)
	ftype edec;//decrease factor
	ftype eterm;//starting epsilon

	int doblock;//0 - no multiple alpha optimization 1 - do multiple alpha optimization
	int blockCmax;
	int blockNmax;
	double blockOver;

	int domaxselection;//WSS-algorithm: 0 - no, 1 - yes, 2 - with selectionfact
	double selectionfact;

	int doreduction;//problem, reduction, reduces precision
	int redsize;
	double redfact;

	//parameters for robust features/supportvectors
	int infold;//for inner ranking, -1=leave-one-out 0,1=no fold >=2 tile in subsets
	int outfold;//for outer crossval, -1=leave-one-out 0,1=no fold >=2 tile in subsets
	//parameters for hyperparameterselection
	ftype ms_e1,ms_e2,ms_ed,ms_c1,ms_c2,ms_cd;
	int flimit;//limit features after training, not deterministic with low C
	int allowCcut;//1: allows cutting the feature ranking with flimit through bounded features (so the result may depend on the feature sequence)
	bool doclass;//modelselection with respect to classification
	bool doregress;//modelselection with respect to regression

	bool doretrain;//retrain features to obtain optimal feature expression levels
	double retrain_fe;//retraining uses a very low epsilon, fe is a factor which reduces used epsilon

	bool dobalanceY;//no bias function scales labels to enforce zero mean
	bool useBalancedErrors;//classification only: use everytime balanced error(tp/2p + tn/2n) instead of standard rate(tp+tn)/(p+n)

	bool useNormLimit;//true: use max and min during normalizationb
};

class Dataset{//stores dyadic matrix data or vectorial data and/or corresponding labels
	public:
	//function
		Dataset();//zero initialization of all members, no memory reserved
		~Dataset();//free all reserved memory
		void free();//free all reserved memory, zero initialization
		void reserve(int n,int dim,bool y);//allocates memory for a dataset of size(n,dim)//y=true: care about y-vector, y=false: delete y-vector//mN and mDim remains unchanged, because the data memory will not be initialized
		void copyFrom(Dataset *src);//copies a complete dataset object from 'src' into 'this'
		bool load(char* fnData,char* fnClass=NULL,int trans=0);//allocates mem, loads data matrix and y-vector from external file with filenames fnData and fnClass, update mN,mDim
		void shuffle(int shufflemode);//random shuffleling of the data objects and labels removes dependence of crossvalidation results with sequence of data objects
		void tileTo(Dataset* learn,Dataset* test,int part,int parts);//generate two new sets 'learn' and 'test' and copy subsets of 'this' into them
		void normalize();//normalizes set 'this' without saving normalization parameters
		void normalize(Dataset *set,bool limit);//normalizes 'this' without saving normalization parameters and //normalizes and limits "set" in the same manner like "this"
		void balance();//scale labels i.o.t. enforce b=0;
	//data
		bool mTrans;//false: vectorial data, stored vector by vector ([obj1, obj2, ...] where obj=[dim1,dim2,...])
				//true: dyadic or kernel data, stored description by description (transposed) [x1(1) x2(1) x3(1)...x1(2) x2(2)...]
		bool mStdKernel;//true: data is pairwise from a kernel - false: data is not from a kernel
		int mN;//currently _used_ number of vectors (data objects)
		int mDim;//currently _used_ number of components per data objects (if dataset is kernelmatrix, this is the number of complex features)
		int mDatamem;//total number of data elements reserved in memory
		int mYmem;//number of y-elements reserved in memory
		ftype *mData; //pointer to x-data
		ftype *mY;//pointer to y-data
};
struct NpEntry{//used in class NormParam
	//normalization-info for one dimension or feature
	//normed_x=(bound(x,min,max)-mean)*scale
	ftype min;//lower limit of datadimension
	ftype max;//upper limit of datadimension
	ftype mean;//mean of datadimension
	ftype scale;//amount of scale
};
class NormParam{//offers functions to apply and remove normalization of kernel matrices or dyadic data
public:
	NormParam();//zero initialization of all members
	~NormParam();//destructor: delete memory
	void free();//delete memory and zero initialization
	void reserve(int dim);//reserves memory for 'dim' normalization parameter records, if nessessary
//	void clear(int dim);//reserve and initialize with zeros
	void getFromSet(Dataset* set);//fills array mX with normalization parameters obtained from set
	void putOnSet(Dataset* set,bool limit);//normalizes "set" with stored normalization parameters mX
	NpEntry *mX;//pointer to array of normalization parameters, mDim valid elements, mDimMem allocated elements
	int mDim;//number of dimensions or features which obtain a normalization 
private:
	int mDimMem;//number of dimensions or features reserved in memory
};

//for block optimiation within SMO
struct sline{//block optimization uses an array of 'sline' and optimized the reduced problem
	int ind;//alpha index
	//input
	ftype f;//f+-epsilon
	ftype a;//for bounding
	ftype low;//low limit (-C or 0)
	ftype high;//high limit (0 or C)
	signed char lowsgn;//low limit as sign(-1 or 0)
	signed char highsgn;//high limit as sign(0 or 1)
	//solution output
	ftype daopt;//solution alphastep
	signed char asgnopt;//solution alpha sign (-1=C,0,1=C,5=free)
};

class SMO
{
	public:
	//functions
		SMO();//constructor: zero initialization of all memory relevant variables
		~SMO();//destructor: release all used memory, result vector mAlpha will also be deleted!
		void free();//release all used memory, result vector mAlpha will also be deleted!
		void run(Dataset *set,ftype b,bool Qreset=true,bool Areset=true);//starts P-SVM feature selection
	//input
		Param *mPar;//pointer to psvm parameters
		ftype mB;//bias in model, mean of set.mY
	//output
		ftype *mAlpha;//will be reserved for same number of elements like input dataset.mDim
		int mFeatures;//number of alphas unequal zero
		//statistics:
		double mTCount; //number of seconds the smo was running
		double mTQCount; //number of seconds the Q-update was running
		double mTBCountsmax;//max single blocktime
		double mTBCounts;//sum blocktime without overhead
		double mTWSS2Count;//total WSS2
		double mTWSS1Count;//total wss1
		double mTU1Count;//time for single updates
		double mTU2Count;//time for double updates
		double mTF1Count;//time for single F updates
		double mTF2Count;//time for double F updates
		double mTBCount;//total block with overhead
		int mACount; //number of iterations
		int mBCount; //number of Block optimizations
		int mQCount;	//number of q-row-updates
		int mBSCount;//number of singular block update tries
		int mBCCount;//number of too complex block update tries
		bool mTimeout; //false: smo complete, true: smo was stopped after timeout
		//status:
		int mShc,mShz;//counts number of indices with alpha=zero(shz) and alpha=+-C(shc)
		double mGeerr,mGeerr1;//(only in debug mode 'check') total error of cost function
		ftype mMaxepsi;//maximum epsilon to find at least one feature (max(fabs(X.'*Y))

		int dd1,dd2;//block optimization statistics

		vector<float> kttlog;//ktt condition (unused)
		vector<int> ann_it;//iterations per annealing step
		vector<double> ann_e;//epsilon per annealing step
		vector<signed char> azlog;//alpha!=0??
		vector<double> alog;//alpha
	private:
//		void init();//initialize Q,F,alpha for set
		inline bool getAlphaByKT(int &aInd);//checking KT-conditions, WSS part 1
		inline void updateQ(int a1Ind);//initialize new row in Q by calculation of scalar products
		inline void getAlphaUpdates(double &da1,double &da2,signed char &al1,signed char &al2,int a1Ind,int a2Ind, ftype a1,ftype f1,signed char l1,signed char u1);//returns update amounts da1,da2 for optimized sup-problem a1Ind,a2Ind
		inline void getAlphaUpdatesnew(double &da1,double &da2,signed char &al1,signed char &al2,int a1Ind,int a2Ind, ftype a1,ftype f1,signed char l1,signed char u1);//alternative(faster?) version:returns update amounts da1,da2 for optimized sup-problem a1Ind,a2Ind
		inline void boxCondition(double &da1,double &da2,signed char &al1,signed char &al2,double temp1,double temp,ftype f1,ftype f2,ftype a1,ftype a2,signed char l1,signed char l2,signed char u1,signed char u2);//calculation of optimal two mAlpha changes with limiting box constraints
		int getdalist(list<sline> *&line);//calculates optimal values for a set of alphas, given in mLine
		int reopt(list<sline> *line,list<sline> *&cc,list<int> &blocks,int lk,int mk);//part of the block update, enforces to test all edges of the hyperbox
		bool solvecholesky(list<sline> *line);//solves mQ*mLine.daopt=mLine.f
		void solveLR(list<sline> *line);//not needed
	//members used in procedures 
		Dataset *mSet;//input dataset reference, filled by run(set)
		int mDim;//input dataset dimensions

		//used for blockoptimization
		double *mMat;//stores submatrix in blockoptimization
		list<sline> *mLine;//stores oszillating indices for blockoptimization
		int mLinecs;//saves number of bounded alphas(mShc) while restarting mLine
		int mLineit;//saves number of iterations(mACount) while restarting mLine

		int mNs;//number of active indices, normally ns=mSet->mDim, in reduction cycles ns=mPar.redsize
		ftype mC;//copy of par->C
		ftype mEpsilon;//decreasing epsilon
		ftype mVio;//maximum violation of KT-Conditions, terminate if vio<epsi*epsitol
		ftype mTerm;

		bool mFirstepsi;//epsilon was just decreased, extra init for mVio
		bool mLastepsi;//epsilon is in final annealing step
		int mLastp;//for doing pseudo random search of next index, for domaxselection=0
		int mRb;//for shifting submatrices in mQ, used in reduction mode
	//internal formula
		ftype **mQ;//matrix kernel^T*kernel, symmetric, positive definite
		ftype *mF;//Q*alpha - data*y
		signed char *mAsgn;//0 - alpha=0; 1 - alpha=+C; -1 - alpha=-C; 5=otherwise
		ftype *mmXY;//-data*y
		int *mSwp;//for swapping indices, used in reduction mode
		ftype **mSubQ;//submatrices of Q, used in reduction mode
		char *mMask;//stores mask of relevant alphas, used in reduction mode

		int mAlphaMem,mQMem,mFMem,mAsgnMem,mmXYMem,mSwpMem,mSubQMem,mMaskMem;//reserved memory
};

class Rankingi
{//stores one feature found by the SMO and offers sorting (std::vector)
	public:
		bool operator<(const Rankingi &a);//local comparator for sort
		ftype a;//mean alpha
		int dimRef;//reference to original dimension
};
bool operator<(const Rankingi &a,const Rankingi &b);//global comparator for sort function
class Ranking
{//stores all features found by the SMO and offers sorting (std::vector)
public:
	vector<Rankingi> mFeat;//informative dimensions of input matrix K
	int mDim;//number of valid dimensions in K (for plausibility check in prediction)
	bool mDeletedByFlimit;//for saving computational costs: true if par.flimit and "allowCcut" is used and more than flimit bounded features occured whithin a hyperparameter selection routine 
};

class MeanRankingi
{//one element of MeanRanking structure, which collects Ranking objects
public:
	bool operator<(const MeanRankingi &b);//local comparison operator
	ftype a;//mean alpha
	int dimRef;//reference to original dimension
	int n;//findings
	int dimRefSum;//mean originalposition (only for sorting)
};
bool operator<(const MeanRankingi &a,const MeanRankingi &b);//global comparison operator

class MeanRanking//offers functionality to average over more than one Ranking
{//used in GetModelWithLoops and GetModelRank(obsolete)
public:
	MeanRanking();//constructor: zero initialization
	~MeanRanking();//destructor: delete all reserved memory
	void free();//release memory and set all members to zero
	void reserve(int dim);//allocate memory for a maximum number of dim features
	void clear(int dim);//allocate memory for dim features, initialize with zero
	void collectFrom(Ranking* rank,int flimit=-1);//adds one Ranking to MeanRanking for averaging
	void extractFromTo(Dataset *from,Dataset* to,int feat=-1);//extract informative dimensions from -> to 
	void transformBack(Ranking *rank);//transforms the rank.dimRefs by this.dimRef
	void putback(Ranking *rank);//stores averaged features of 'this' into Ranking 'rank' for building psvm model
	void copyFrom(MeanRanking* mrank);//copies a whole MeanRanking object from 'mrank' into 'this'
	MeanRankingi* mrFeat;//pointer to array with collected features
	int mDimMem;//max nr of features
	int mCollectedFeatures;//number of collected features
	int mCollectedSets;//number of collected sets
	int mDim;//number of dimensions used in feature extraction
	double mEpsilon;//hyperparameter used to get the ranking
	double mC;
	bool mDeletedByFlimit;//for saving computational costs: true if par.flimit and "allowCcut" is used and more than flimit bounded features occured whithin a hyperparameter selection routine 
};

class rs
{//every element stores one ROC point
public:
	bool operator<(const rs &b);//local comparison operator
	ftype yo;
	ftype yt;
};
bool operator<(const rs &a,const rs &b);//global comparator for sort function

struct Msdata
{//hyperparameter selection grid point
	double c,e;//hyperparams
	double eec,eer;//errors
};
struct Trainres{
	double c_min,c_max,e_min,e_max;//hyperparameter search area
	int n_hyp; //number of searched hyperparameters

	double reg_Gc,reg_Ge,reg_TError,reg_GError;//Hyperparameters,TrainError,GeneralizationError
	double reg_Tc,reg_Te;//sometimes mean
	double cla_Gc,cla_Ge,cla_TError,cla_GError;
	double cla_Tc,cla_Te;

	int reg_Tnf,reg_Gnf;//(mean) features
	int cla_Tnf,cla_Gnf;//(mean) features

	//following are currently not used
	double reg_emin,reg_emax,reg_emean,reg_cmin,reg_cmax,reg_cmean;//hyperparameter-finding area
	double cla_emin,cla_emax,cla_emean,cla_cmin,cla_cmax,cla_cmean;//hyperparameter-finding area
	//following four results are now used!!!
	int cla_Ttp,cla_Ttn,cla_Tfp,cla_Tfn;//number of miss classifications
//	double cla_Tacc/*,cla_Taccb*/;//accuracy,balanced accuracy
//	double cla_Gacc/*,cla_Gaccb*/;//accuracy,balanced accuracy
//	double cla_Terr/*,cla_Terrb*/;//error,balanced error
//	double cla_Gerr/*,cla_Gerrb*/;//error,balanced error
	int cla_Gtp,cla_Gtn,cla_Gfp,cla_Gfn;//number of miss classifications

	//vector<vector<int>> cvlist;//list of features obtained in single cv-tasks
	
//	MeanRanking reg_mrank;//not used, the content is currently invalid
//	MeanRanking cla_mrank;//not used, the content is currently invalid
	double maxFeat;//only for user information: maximum of lagrangian alpha, helpful to select C because C limits alpha. If maxFeat==C, increasing C can change the predictor, otherwise not

	int msne;//number of tested epsilon parameters in ms
	int msnc;//number of tested Cparameters in ms
//	vector<Msdata> ms;
	Msdata *ms;
};
class PSVMmodel{
public:
	PSVMmodel();//constructor: clear all members
	~PSVMmodel();//destructor:free used memory
	//model operations
	bool save(char* fn);//saves b, np, ranking to a file with name 'fn'
	bool load(char* fn);//loads b, np, ranking from a file with name 'fn', returns false, if file not found
	void free();//clear model components and release memory
	//dataset operations
	void kernel(Dataset* xset,Dataset* zset);//calculate kernel matrix k(x,z) into xset
	void kernel(Dataset* xset);//calculate kernel matrix k(x,x) or k(x) into xset

	//using inline kernel functions is faster but its not possible to use dynamic function pointer
	//direct kernel fuctions
	inline double kf_LINEAR(double x);
	inline double kf_POLY(double x);
	inline double kf_RBF(double x);
	inline double kf_SIGMOID(double x);
	inline double kf_SIN(double x);
	inline double kf_PLUMMER(double x);
	inline double kf_USER(double x);
	//preparing the vector distance for kernel function
	inline ftype kf_LINEAR(ftype* x,ftype* z,int n);
	inline ftype kf_POLY(ftype* x,ftype* z,int n);
	inline ftype kf_RBF(ftype* x,ftype* z,int n);
	inline ftype kf_SIGMOID(ftype* x,ftype* z,int n);
	inline ftype kf_SIN(ftype* x,ftype* z,int n);
	inline ftype kf_PLUMMER(ftype* x,ftype* z,int n);
	inline ftype kf_USER(ftype* x,ftype* z,int n);

	//model generation
	bool getRankFromSet(Dataset* set,bool Qreset=true,bool Areset=true);//run smo and get feature ranking in "rank" (sparse vector alpha)
		//feature limitation: model stores all features, but uses only the allowed number of features
	void getB(Dataset* set);//get model parameter "b" for "set"
	void getSVZ(Dataset* set);//requires valid 'rank' and saves selected support vectors into svz
	bool getModelWithLoops(Dataset* set,Trainres *res,bool getmodel=true);//uses hyperparameter selection and or crossvalidation cycles for training a predictor for dataset 'set'
		//attention!!!: in crossvalidation mode the final model is not valid because it belongs to the last subset training, a retraining over whole set is required
		//returns false if at least one SMO training phase was stopped after timeout.
		//feature limitation: model stores all features, but uses only the allowed number of features
	bool getModel(Dataset* set,Trainres *res,bool Qreset=true,bool Areset=true);//gets model for kernelmatrix set, puts results into res, normalizes set!!!
//	bool getModelRank(Dataset* set,Trainres *res);//experimental status for building an average feature ranking (results are not as good as getModelWithLoops

	//using the model
	void getY(ftype* yy,Dataset* set,int flimit=-1,bool full=true);//computes Y for existing model using
	//	- rank[0..limitfeat-1]; if limitfeat ==-1 use all rank-elements
	//	- b (level shift)
	//	- np (normalization) if full==true
	//	- svz (build relations with support features) if full == true && !isrelation
	//	caller is responsible for allocating memory y

	//	ftype getSingleY(ftype* kvector,int flimit=-1);//uses kernel and supportvectors if not relational, normalize
	//	ftype getSingleY_direct(ftype* kvector,int flimit=-1);//no kernel no normalization
	//	void addB(ftype* kvector);
//	void getErrors(ftype* y,Dataset* ms,double *mse,double* rate);//gets mean squared error and classification rate by comparing y with ms.mY
	//void getAddErrors(ftype* y,Dataset* ms,double *mseadd,double *rateadd,double *mse,double* rate);//gets MSE and classification error of y-vector vs. dataset.mY, cummulates mseadd and rateadd
	void getErrors(ftype* y,Dataset* ms,double *mse,double* rate,int* tp, int* tn, int* fp, int* fn);
	void getAddErrors(ftype* y,Dataset* ms,double *mseadd,/* double *rateadd,*/double *mse,double* rate,int* tpa,int* tna,int* fpa,int* fna);
	double getCError(double tp,double tn,double fp,double fn);
//	void getAddErrors(ftype* y,Dataset* ms,double *mseadd, double *rateadd,double *mse,double* rate,int* tpa,int* tna,int* fpa,int* fna,int* tp,int* tn,int* fp,int* fn);
	//Model output, managed memory
	//the model is at least one of (rank, np, b, svz)
	NormParam np;//filled by getNormParam, load, reserved for 0..dim-1, even if sparse 
	ftype b;//scalar
	Ranking rank;//compressed alpha as sorted feature ranking, maximum dimRef is dim-1
	Dataset *svz;//support vectors(1..N,1..rank.feat.size), only in nonrelational mode, see par.isrelation
	Param par;//Parameter input, filled with defaults by constructor
	SMO smo;
};

extern bool verbose;//true: print additional progress messages
extern bool verbimp;//true: print important messages
extern bool verb_ms;
extern FILE* flog;
